from typing import Any, Optional, Union

from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext

UnionType = type(Union[int, str])


class OptionalFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) is UnionType and len(annotation.__args__) == 2 and annotation.__args__[1] is type(None):
            checker = ctx.find_checker(annotation.__args__[0])
            if checker is None:
                return None
            else:
                return OptionalChecker(checker)
        else:
            return None


class OptionalChecker(TypeChecker):
    inner: TypeChecker

    def __init__(self, inner: TypeChecker):
        self.inner = inner

    def check_and_wrap(self, arg: Any, upper: ExecutionContext) -> Any:
        if arg is None:
            return arg
        else:
            ctx = OptionalExecutionContext(upper, [self.inner], 0)
            return self.inner.check_and_wrap(arg, ctx)

    def describe(self) -> str:
        return f"Optional[{self.inner.describe()}]"

    def base_type(self) -> list[Any]:
        return [self.inner.base_type()]


class OptionalExecutionContext(CompoundTypeExecutionContext):
    def name(self):
        return "Optional"



import ast
import importlib
from collections.abc import Callable
from importlib.abc import MetaPathFinder
from importlib.machinery import SourceFileLoader
from importlib.util import decode_source


def install_import_hook(should_patch_predicate: Callable[[str], bool],
                        transformer: Callable[[str], ast.NodeTransformer]):
    import sys

    already_patched = next((f for f in sys.meta_path if isinstance(f, UntypyFinder)), None)
    if already_patched is not None:
        return

    original_finder = next(f for f in sys.meta_path if f.__name__ == 'PathFinder' and hasattr(f, 'find_spec'))
    sys.meta_path.insert(0, UntypyFinder(original_finder, should_patch_predicate, transformer))


class UntypyFinder(MetaPathFinder):

    def __init__(self, inner_finder: MetaPathFinder, should_patch_predicate: Callable[[str], bool],
                 transformer: Callable[[str], ast.NodeTransformer]):
        self.inner_finder = inner_finder
        self.should_patch_predicate = should_patch_predicate
        self.transformer = transformer

    def find_spec(self, fullname, path=None, target=None):
        if not self.should_instrument(fullname):
            return None

        inner_spec = self.inner_finder.find_spec(fullname, path, target)
        if inner_spec is not None and isinstance(inner_spec.loader, SourceFileLoader):
            inner_spec.loader = UntypyLoader(inner_spec.loader.name, inner_spec.loader.path, self.transformer)
        return inner_spec

    def should_instrument(self, module_name: str) -> bool:
        return self.should_patch_predicate(module_name)


class UntypyLoader(SourceFileLoader):

    def __init__(self, fullname, path, transformer: Callable[[str], ast.NodeTransformer]):
        super().__init__(fullname, path)
        self.transformer = transformer

    def source_to_code(self, data, path, *, _optimize=-1):
        source = decode_source(data)
        tree = compile(source, path, 'exec', ast.PyCF_ONLY_AST,
                       dont_inherit=True, optimize=_optimize)
        self.transformer(self.name.split('.')).visit(tree)
        ast.fix_missing_locations(tree)
        return compile(tree, path, 'exec', dont_inherit=True, optimize=_optimize)

    def exec_module(self, module) -> None:
        # cache_from_source has to be patched to prevent load from cache
        # this enables patching of AST
        # See https://github.com/agronholm/typeguard/blob/89c1478bd33bcf9a7cccc2c962ebeaa034e51908/src/typeguard/importhook.py#L12
        original = getattr(importlib._bootstrap_external, 'cache_from_source')
        try:
            setattr(importlib._bootstrap_external, 'cache_from_source', lambda *args: None)
            return super().exec_module(module)
        finally:
            setattr(importlib._bootstrap_external, 'cache_from_source', original)



import collections.abc
import inspect
import sys
from collections.abc import Generator
from typing import Any, Optional
from typing import Generator as OtherGenerator

from untypy.error import UntypyTypeError, UntypyAttributeError, Location
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext, NoResponsabilityWrapper

GeneratorTypeA = type(Generator[None, None, None])
GeneratorTypeB = type(OtherGenerator[None, None, None])


class GeneratorFactory(TypeCheckerFactory):
    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) in [GeneratorTypeA, GeneratorTypeB] and annotation.__origin__ == collections.abc.Generator:
            if len(annotation.__args__) != 3:
                raise ctx.wrap(UntypyAttributeError(f"Expected 3 type arguments for Generator."))

            (yield_checker, send_checker, return_checker) = list(
                map(lambda a: ctx.find_checker(a), annotation.__args__))

            if yield_checker is None:
                raise ctx.wrap(UntypyAttributeError(f"The Yield Annotation of the Generator could not be resolved."))
            if send_checker is None:
                raise ctx.wrap(UntypyAttributeError(f"The Send Annotation of the Generator could not be resolved."))
            if return_checker is None:
                raise ctx.wrap(UntypyAttributeError(f"The Return Annotation of the Generator could not be resolved."))

            return GeneratorChecker(yield_checker, send_checker, return_checker)
        else:
            return None


class GeneratorChecker(TypeChecker):
    yield_checker: TypeChecker
    send_checker: TypeChecker
    return_checker: TypeChecker

    def __init__(self, yield_checker: TypeChecker, send_checker: TypeChecker, return_checker: TypeChecker):
        self.yield_checker = yield_checker
        self.send_checker = send_checker
        self.return_checker = return_checker

    def may_be_wrapped(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not inspect.isgenerator(arg):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        me = self
        yield_ctx = TypedGeneratorYieldReturnContext(arg, self, True, ctx)
        return_ctx = TypedGeneratorYieldReturnContext(arg, self, False, ctx)

        def wrapped():
            try:
                sent = None
                while True:
                    value_yield = arg.send(sent)
                    # check value_yield (arg is responsable)
                    value_yield = me.yield_checker.check_and_wrap(value_yield, yield_ctx)

                    sent = yield value_yield

                    caller = sys._getframe(1)

                    # check sent value (caller is responsable)
                    sent = me.send_checker.check_and_wrap(sent, TypedGeneratorSendContext(caller, me, ctx))

            except StopIteration as e:
                # check value_returned (arg is responsable)
                ret = me.return_checker.check_and_wrap(e.value, return_ctx)
                return ret

        return wrapped()

    def describe(self) -> str:
        return f"Generator[{self.yield_checker.describe()}, {self.send_checker.describe()}, {self.return_checker.describe()}]"

    def base_type(self) -> Any:
        return [GeneratorType]


class TypedGeneratorYieldReturnContext(CompoundTypeExecutionContext):
    generator: Generator[Any, Any, Any]

    def __init__(self, generator: Generator[Any, Any, Any], checker: GeneratorChecker, is_yield: bool,
                 upper: ExecutionContext):
        self.generator = generator
        # index in checkers list
        if is_yield:
            idx = 0
        else:
            idx = 2
        super().__init__(upper, [checker.yield_checker, checker.send_checker, checker.return_checker], idx)

    def name(self) -> str:
        return "Generator"

    def responsable(self) -> Optional[Location]:
        try:
            if hasattr(self.generator, 'gi_frame'):
                return Location(
                    file=inspect.getfile(self.generator.gi_frame),
                    line_no=inspect.getsourcelines(self.generator.gi_frame)[1],
                    source_line="\n".join(inspect.getsourcelines(self.generator.gi_frame)[0]),
                )
        except OSError:  # this call does not work all the time
            pass
        except TypeError:
            pass
        return None


class TypedGeneratorSendContext(CompoundTypeExecutionContext):
    def __init__(self, stack: inspect.FrameInfo, checker: GeneratorChecker, upper: ExecutionContext):
        self.stack = stack
        super().__init__(NoResponsabilityWrapper(upper),
                         [checker.yield_checker, checker.send_checker, checker.return_checker], 1)

    def name(self) -> str:
        return "Generator"

    def responsable(self) -> Optional[Location]:
        return Location.from_stack(self.stack)



from typing import Any, Optional, Sequence
from collections.abc import Sequence as ABCSequence

from untypy.error import UntypyTypeError, UntypyAttributeError
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext
from untypy.util import CompoundTypeExecutionContext
from untypy.impl.simple import SimpleFactory
from untypy.impl.list import ListChecker
from untypy.impl.tuple import VariadicTupleChecker
from untypy.impl.union import UnionChecker

SequenceTypeA = type(Sequence[int])
SequenceTypeB = type(ABCSequence[int])


class SequenceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        t = type(annotation)
        if (t in [SequenceTypeA, SequenceTypeB] and annotation.__origin__ in [Sequence, ABCSequence]) or \
                annotation in [Sequence, ABCSequence]:  # no args version
            try:
                args = annotation.__args__
            except AttributeError:
                args = []
            inner = []
            elemChecker = None
            if len(args) == 0:
                sf = SimpleFactory()
                inner = [sf.create_from(list, ctx),
                         sf.create_from(tuple, ctx),
                         sf.create_from(str, ctx)]
            elif len(args) == 1:
                elemChecker = ctx.find_checker(args[0])
                if elemChecker is None:
                    return None
                inner = [ListChecker(elemChecker, ctx.declared_location()),
                         VariadicTupleChecker(elemChecker)]
            return SequenceChecker(inner, ctx, elemChecker)
        else:
            return None


class SequenceChecker(UnionChecker):

    elemChecker: Optional[TypeChecker]

    def __init__(self, inner: list[TypeChecker], ctx: CreationContext, elemChecker: Optional[TypeChecker]):
        super().__init__(inner, ctx)
        self.elemChecker = elemChecker

    def describe(self) -> str:
        if self.elemChecker:
            desc = self.elemChecker.describe()
            return f"Sequence[{desc}]"
        else:
            return "Sequence"



from __future__ import annotations

import inspect
from enum import Enum
from typing import Any, Optional, Tuple, Iterable


class Location:
    file: str
    line_no: int
    source_line: str

    def __init__(self, file: str, line_no: int, source_line: str):
        self.file = file
        self.line_no = line_no
        self.source_line = source_line

    def __str__(self):
        buf = f"{self.file}:{self.line_no}"
        if self.source_line:
            for i, line in enumerate(self.source_line.splitlines()):
                if i < 5:
                    buf += f"\n{'{:3}'.format(self.line_no + i)} | {line}"
            if i >= 5:
                buf += "\n    | ..."
        return buf

    def __repr__(self):
        return f"Location(file={self.file.__repr__()}, line_no={self.line_no.__repr__()}, source_line={repr(self.source_line)})"

    def __eq__(self, other):
        if not isinstance(other, Location):
            return False
        return self.file == other.file and self.line_no == other.line_no

    @staticmethod
    def from_code(obj) -> Location:
        try:
            return Location(
                file=inspect.getfile(obj),
                line_no=inspect.getsourcelines(obj)[1],
                source_line="".join(inspect.getsourcelines(obj)[0]),
            )
        except Exception:
            return Location(
                file=inspect.getfile(obj),
                line_no=1,
                source_line=repr(obj)
            )

    @staticmethod
    def from_stack(stack) -> Location:
        if isinstance(stack, inspect.FrameInfo):
            try:
                return Location(
                    file=stack.filename,
                    line_no=stack.lineno,
                    source_line=stack.code_context[0]
                )
            except Exception:
                return Location(
                    file=stack.filename,
                    line_no=stack.lineno,
                    source_line=None
                )
        else:  # assume sys._getframe(...)
            try:
                source_line = inspect.findsource(stack.f_code)[0][stack.f_lineno - 1]
                return Location(
                    file=stack.f_code.co_filename,
                    line_no=stack.f_lineno,
                    source_line=source_line
                )
            except Exception:
                return Location(
                    file=stack.f_code.co_filename,
                    line_no=stack.f_lineno,
                    source_line=None
                )


class Frame:
    type_declared: str
    indicator_line: str

    declared: Optional[Location]
    responsable: Optional[Location]

    responsibility_type: Optional[ResponsibilityType]

    def __init__(self, type_declared: str, indicator_line: Optional[str],
                 declared: Optional[Location], responsable: Optional[Location]):

        self.type_declared = type_declared
        if indicator_line is None:
            indicator_line = '^' * len(type_declared)
        self.indicator_line = indicator_line
        self.declared = declared
        self.responsable = responsable

    def __str__(self):
        buf = f"in: {self.type_declared}\n" \
              f"    {self.indicator_line}\n"

        if self.responsable is not None:
            buf += f"{self.responsable.file}:{self.responsable.line_no}:\n" \
                   f"{self.responsable.source_line}\n" \
                   f"\n"
        return buf


class ResponsibilityType(Enum):
    IN = 0
    OUT = 1

    def invert(self):
        if self is ResponsibilityType.IN:
            return ResponsibilityType.OUT
        else:
            return ResponsibilityType.IN

def joinLines(l: Iterable[str]) -> str:
    return '\n'.join([x.rstrip() for x in l])

# Note: the visual studio code plugin uses the prefixes "caused by: " and "declared at: "
# for finding source locations. Do not change without changing the plugin code!!
CAUSED_BY_PREFIX = "caused by: "
DECLARED_AT_PREFIX = "declared at: "

def formatLocations(prefix: str, locs: list[Location]) -> str:
    return joinLines(map(lambda s: prefix + str(s), locs))

class UntypyTypeError(TypeError):
    given: Any
    expected: str
    frames: list[Frame]
    notes: list[str]
    previous_chain: Optional[UntypyTypeError]
    responsibility_type: ResponsibilityType

    def __init__(self, given: Any, expected: str, frames: list[Frame] = [],
                 notes: list[str] = [],
                 previous_chain: Optional[UntypyTypeError] = None,
                 responsibility_type: ResponsibilityType = ResponsibilityType.IN):
        self.responsibility_type = responsibility_type
        self.given = given
        self.expected = expected
        self.frames = frames.copy()
        for frame in self.frames:
            if frame.responsibility_type is None:
                frame.responsibility_type = responsibility_type
        self.notes = notes.copy()
        self.previous_chain = previous_chain
        super().__init__('\n' + self.__str__())

    def next_type_and_indicator(self) -> Tuple[str, str]:
        if len(self.frames) >= 1:
            frame = self.frames[-1]
            return frame.type_declared, frame.indicator_line
        else:
            return self.expected, "^" * len(self.expected)

    def with_frame(self, frame: Frame) -> UntypyTypeError:
        frame.responsibility_type = self.responsibility_type
        return UntypyTypeError(self.given, self.expected, self.frames + [frame],
                               self.notes, self.previous_chain, self.responsibility_type)

    def with_previous_chain(self, previous_chain: UntypyTypeError):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes, previous_chain, self.responsibility_type)

    def with_note(self, note: str):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes + [note], self.previous_chain, self.responsibility_type)

    def with_inverted_responsibility_type(self):
        return UntypyTypeError(self.given, self.expected, self.frames,
                               self.notes, self.previous_chain, self.responsibility_type.invert())

    def last_responsable(self):
        for f in reversed(self.frames):
            if f.responsable is not None and f.responsibility_type is ResponsibilityType.IN:
                return f.responsable
        return None

    def last_declared(self):
        for f in reversed(self.frames):
            if f.declared is not None:
                return f.declared
        return None

    def __str__(self):
        declared_locs = []
        responsable_locs = []

        for f in self.frames:
            if f.responsable is not None and f.responsibility_type is ResponsibilityType.IN and str(
                    f.responsable) not in responsable_locs:
                responsable_locs.append(str(f.responsable))
            if f.declared is not None and str(f.declared) not in declared_locs:
                declared_locs.append(str(f.declared))

        cause = formatLocations(CAUSED_BY_PREFIX, responsable_locs)
        declared = formatLocations(DECLARED_AT_PREFIX, declared_locs)

        (ty, ind) = self.next_type_and_indicator()

        notes = joinLines(self.notes)
        if notes:
            notes = notes + "\n\n"

        if self.previous_chain is None:
            previous_chain = ""
        else:
            previous_chain = self.previous_chain.__str__()
        if previous_chain:
            previous_chain = previous_chain.strip() + "\n\n"

        ctx = ""
        if self.expected != ty:
            ctx = f"context: {ty.rstrip()}\n" \
                  f"         {ind.rstrip()}"

        given = repr(self.given)
        expected = self.expected.strip()
        if expected != 'None':
            expected = f'value of type {expected}'
        return (f"""
{previous_chain}{notes}given:    {given.rstrip()}
expected: {expected}

{ctx}
{declared}

{cause}""")


class UntypyAttributeError(AttributeError):

    def __init__(self, message: str, locations: list[Location] = []):
        self.message = message
        self.locations = locations.copy()
        super().__init__(self.__str__())

    def with_location(self, loc: Location) -> UntypyAttributeError:
        return type(self)(self.message, self.locations + [loc]) # preserve type

    def __str__(self):
        return f"{self.message}\n{formatLocations(DECLARED_AT_PREFIX, self.locations)}"


class UntypyNameError(UntypyAttributeError):
    pass



from collections.abc import Iterator, Iterable
from typing import TypeVar, Optional, Any, Generic, Dict, List, Set

from untypy.error import UntypyAttributeError, UntypyTypeError
from untypy.impl.protocol import ProtocolChecker
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import TypeCheckerFactory, TypeChecker, CreationContext, ExecutionContext
from untypy.util import ReplaceTypeExecutionContext

K = TypeVar("K")
V = TypeVar("V")


class WDict(Generic[K, V], dict):
    def __getitem__(self, item: K) -> V:
        pass

    def __setitem__(self, key: K, value: V) -> None:
        pass


I = TypeVar("I")


class WList(Generic[I], list):
    def __getitem__(self, item: int) -> I:
        pass

    def __setitem__(self, key: int, value: I) -> None:
        pass


I = TypeVar("I")


class WSet(Generic[I], set):
    def __contains__(self, item: I) -> bool:
        pass

    def add(self, elem: I) -> None:
        pass


I = TypeVar("I")


class WIterable(Generic[I]):
    def __iter__(self) -> Iterator[I]:
        pass

InterfaceMapping = {
    dict: (WDict,),
    Dict: (WDict,),
    list: (WList,),
    List: (WList,),
    set: (WSet,),
    Set: (WSet,),
    Iterable: (WIterable,),
}


class InterfaceFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if hasattr(annotation, '__origin__') and hasattr(annotation,
                                                         '__args__') and annotation.__origin__ in InterfaceMapping:
            (protocol,) = InterfaceMapping[annotation.__origin__]
            bindings = protocol.__parameters__  # args of Generic super class
            origin = annotation.__origin__

            inner_checkers = []
            for param in annotation.__args__:
                ch = ctx.find_checker(param)
                if ch is None:
                    raise UntypyAttributeError(f"Could not resolve annotation {param} inside of {annotation}")
                inner_checkers.append(ch)
            if len(inner_checkers) != len(bindings):
                raise UntypyAttributeError(f"Expected {len(bindings)} type arguments inside of {annotation}")

            name = f"{origin.__name__}[" + (', '.join(map(lambda t: t.describe(), inner_checkers))) + "]"

            bindings = dict(zip(bindings, annotation.__args__))
            ctx = ctx.with_typevars(bindings)
            if type(origin) == type:
                template = WrappedType(protocol, ctx.with_typevars(bindings), name=name, implementation_template=origin,
                                       declared=ctx.declared_location())
                return InterfaceChecker(origin, template, name)
            else:
                # type(origin) == collection.abc.ABCMeta
                return ProtocolChecker(protocol, ctx, altname=name)

        else:
            return None


class InterfaceChecker(TypeChecker):

    def __init__(self, origin, template, name):
        self.origin = origin
        self.template = template
        self.name = name
        pass

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not issubclass(type(arg), self.origin):
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

        instance = self.template.__new__(self.template)
        instance._WrappedClassFunction__inner = arg
        instance._WrappedClassFunction__return_ctx = ReplaceTypeExecutionContext(ctx, self.name)
        return instance

    def describe(self) -> str:
        return self.name



import ast
import inspect
import sys
from types import ModuleType
from typing import Optional, Any, Union

from .patching import wrap_function, patch_class, wrap_class, DefaultConfig
from .patching.ast_transformer import UntypyAstTransformer, did_no_code_run_before_untypy_enable, \
    UntypyAstImportTransformer
from .patching.import_hook import install_import_hook
from .util.condition import FunctionCondition

GlobalConfig = DefaultConfig


def just_install_hook(prefixes=[]):
    def predicate(module_name):
        for p in prefixes:
            if module_name == p:
                return True
            elif module_name.startswith(p + "."):
                return True
        return False

    install_import_hook(predicate, lambda path: UntypyAstTransformer())


def just_transform(source, modname, symbol='exec'):
    tree = compile(source, modname, symbol, flags=ast.PyCF_ONLY_AST, dont_inherit=True, optimize=-1)
    transform_tree(tree)
    return tree

def transform_tree(tree):
    UntypyAstTransformer().visit(tree)
    ast.fix_missing_locations(tree)

def enable(*, recursive: bool = True, root: Union[ModuleType, str, None] = None, prefixes: list[str] = []) -> None:
    global GlobalConfig
    caller = _find_calling_module()
    exit_after = False

    if root is None:
        root = caller
        exit_after = True
    if caller is None:
        raise Exception("Couldn't find loading Module. This is a Bug.")

    rootname = root
    if hasattr(rootname, '__name__'):
        rootname = root.__name__

    GlobalConfig = DefaultConfig._replace(checkedprefixes=[rootname])

    def predicate(module_name):
        if recursive:
            # Patch if Submodule
            if module_name.startswith(rootname + "."):
                return True
            else:
                for p in prefixes:
                    if module_name == p:
                        return True
                    elif module_name.startswith(p + "."):
                        return True
            return False
        else:
            raise AssertionError("You cannot run 'untypy.enable()' twice!")

    transformer = lambda path: UntypyAstTransformer()
    install_import_hook(predicate, transformer)
    _exec_module_patched(root, exit_after, transformer(caller.__name__.split(".")))


def enable_on_imports(*prefixes):
    print("!!!! THIS FEATURE HAS UNEXPECTED SIDE EFFECTS WHEN IMPORTING ABSOLUTE SUBMODULES !!!")
    # TODO: Fix import of submodules should not change parent module.
    global GlobalConfig
    GlobalConfig = DefaultConfig._replace(checkedprefixes=[*prefixes])
    caller = _find_calling_module()

    def predicate(module_name: str):
        module_name = module_name.replace('__main__.', '')
        for p in prefixes:
            if module_name == p:
                return True
            elif module_name.startswith(p + "."):
                return True
            else:
                return False

    transformer = lambda path: UntypyAstImportTransformer(predicate, path)
    install_import_hook(predicate, transformer)
    _exec_module_patched(caller, True, transformer(caller.__name__.split(".")))


def _exec_module_patched(mod: ModuleType, exit_after: bool, transformer: ast.NodeTransformer):
    source = inspect.getsource(mod)
    tree = compile(source, mod.__file__, 'exec', ast.PyCF_ONLY_AST,
                   dont_inherit=True, optimize=-1)

    if not did_no_code_run_before_untypy_enable(tree):
        raise AssertionError("Please put 'untypy.enable()' at the start of your module like so:\n"
                             "\timport untypy\n"
                             "\tuntypy.enable()")

    transformer.visit(tree)
    ast.fix_missing_locations(tree)
    patched_mod = compile(tree, mod.__file__, 'exec', dont_inherit=True, optimize=-1)
    stack = list(map(lambda s: s.frame, inspect.stack()))
    try:
        exec(patched_mod, mod.__dict__)
    except Exception as e:
        while e.__traceback__.tb_frame in stack:
            e.__traceback__ = e.__traceback__.tb_next
        sys.excepthook(type(e), e, e.__traceback__)
        if exit_after: sys.exit(-1)
    if exit_after: sys.exit(0)


def _find_calling_module() -> Optional[ModuleType]:
    for caller in inspect.stack():
        if caller.filename != __file__:
            mod = inspect.getmodule(caller.frame)
            if mod is not None:
                return mod
    return None


def _condgetfc(func):
    if hasattr(func, "__fc"):
        return getattr(func, "__fc")
    else:
        fc = FunctionCondition()
        setattr(func, "__fc", fc)
        fc.func = func
        return fc


def precondition(cond):
    def decorator(func):
        fc = _condgetfc(func)
        fc.precondition.append(cond)
        return func

    return decorator


def postcondition(cond):
    def decorator(func):
        fc = _condgetfc(func)
        fc.postcondition.append(cond)
        return func

    return decorator


def unchecked(fn):
    setattr(fn, "__unchecked", True)
    return fn


def patch(a: Any) -> Any:
    global GlobalConfig
    if hasattr(a, '__unchecked'):
        return a

    if inspect.isfunction(a):
        return wrap_function(a, GlobalConfig)
    elif inspect.isclass(a):
        patch_class(a, GlobalConfig)
        return a
    else:
        return a


typechecked = patch


def wrap_import(a: Any) -> Any:
    global GlobalConfig
    if inspect.isfunction(a):
        return wrap_function(a, GlobalConfig)
    elif inspect.isclass(a) or inspect.ismodule(a):
        return wrap_class(a, GlobalConfig)
    else:
        return a



import inspect
import re
from collections.abc import Callable
from typing import Optional

from untypy.error import UntypyAttributeError, UntypyTypeError, Frame
from untypy.interfaces import ExecutionContext, WrappedFunction

WrappedFunctionContextProvider = Callable[[str], ExecutionContext]


class FunctionCondition:

    def __init__(self):
        self.precondition = []
        self.postcondition = []
        self.func = None
        pass

    def prehook(self, boundargs, ctx: WrappedFunctionContextProvider):
        for p in self.precondition:
            bindings = {}
            for name in inspect.signature(p).parameters:
                if name in boundargs.arguments:
                    bindings[name] = boundargs.arguments[name]
                else:
                    raise UntypyAttributeError(
                        f"Did not find argument {name} of precondition in function.",
                        locations=[
                            WrappedFunction.find_location(p),
                            WrappedFunction.find_location(self.func),
                        ]
                    )
            if not p(**bindings):
                lsource = find_lambdasource(p)
                if lsource is not None:
                    expected = f"passing: {lsource}"
                else:
                    expected = "passing precondition"

                err = UntypyTypeError(
                    bindings,
                    expected
                )
                err = err.with_note("Failed precondition.")
                err = err.with_frame(Frame(
                    expected,
                    "",
                    declared=WrappedFunction.find_location(p),
                    responsable=None,
                ))
                raise ctx(0).wrap(err)

    def posthook(self, ret, boundargs, ctx: ExecutionContext):
        for p in self.postcondition:
            bindings = {}
            for name in inspect.signature(p).parameters:
                if name == "ret":
                    bindings["ret"] = ret
                elif name in boundargs.arguments:
                    bindings[name] = boundargs.arguments[name]
                else:
                    raise UntypyAttributeError(
                        f"Did not find argument {name} of postcondition in function.",
                        locations=[
                            WrappedFunction.find_location(p),
                            WrappedFunction.find_location(self.func),
                        ]
                    )
            if not p(**bindings):
                lsource = find_lambdasource(p)
                if lsource is not None:
                    expected = f"passing: {lsource}"
                else:
                    expected = "passing postcondition"

                given = ret.__repr__()
                err = UntypyTypeError(
                    given,
                    expected,
                ).with_note("Failed postcondition").with_frame(Frame(
                    expected,
                    "",
                    declared=WrappedFunction.find_location(p),
                    responsable=None,
                ))
                raise ctx.wrap(err)


def find_lambdasource(fn) -> Optional[str]:
    """
    tries to retuns body of precondition or postcondition annotation
    """
    try:
        fn = WrappedFunction.find_original(fn)
        source = inspect.getsource(fn).split('\n')
        m = re.match('@[a-zA-Z_\.]+\((.*)\)', source[0])
        if m is not None and len(m.groups()) == 1:
            return m.group(1)
    except:
        return None



from typing import Any, Optional, Literal

from untypy.error import UntypyTypeError
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext

LiteralType = type(Literal[42])


class LiteralFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if type(annotation) is LiteralType and annotation.__origin__ == Literal:
            return LiteralChecker(annotation.__args__)
        else:
            return None


class LiteralChecker(TypeChecker):
    inner: list[Any]

    def __init__(self, inner: list[Any]):
        self.inner = inner

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if arg in self.inner:
            return arg
        else:
            raise ctx.wrap(UntypyTypeError(
                arg,
                self.describe()
            ))

    def base_type(self) -> list[Any]:
        return self.inner[:]

    def describe(self) -> str:
        strings = map(lambda v: "%r" % v, self.inner)
        return f"Literal[{', '.join(strings)}]"



import typing
from typing import Optional, TypeVar, Any

from untypy.error import UntypyTypeError
from untypy.impl.protocol import ProtocolChecker
from untypy.interfaces import TypeCheckerFactory, CreationContext, TypeChecker, ExecutionContext


class GenericProtocolChecker(ProtocolChecker):
    def protocol_type(self) -> str:
        return "Generic"

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if not isinstance(arg, self.proto):
            raise ctx.wrap(UntypyTypeError(
                expected=self.describe(),
                given=arg
            )).with_note(f"Type '{type(arg).__name__}' does not inherit from '{self.proto.__name__}'")
        return super().check_and_wrap(arg, ctx)


class GenericFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        # TODO: Support other typevar features
        if type(annotation) is TypeVar:
            (found, replacement_annotation) = ctx.resolve_typevar(annotation)
            if found:
                inner = ctx.find_checker(replacement_annotation)
                if inner is not None:
                    return BoundTypeVar(inner, annotation)
                else:
                    return None
            else:
                return UnboundTypeVar(annotation)
        elif hasattr(annotation, '__args__') and hasattr(annotation.__origin__,
                                                         '__mro__') and typing.Generic in annotation.__origin__.__mro__:
            return GenericProtocolChecker(annotation, ctx)
        else:
            return None


class BoundTypeVar(TypeChecker):
    def __init__(self, inner: TypeChecker, typevar: TypeVar):
        self.inner = inner
        self.typevar = typevar

    def describe(self) -> str:
        return f"{self.typevar}={self.inner.describe()}"

    def may_be_wrapped(self) -> bool:
        return self.inner.may_be_wrapped()

    def base_type(self) -> list[Any]:
        return self.inner.base_type()

    def base_type_priority(self) -> int:
        return self.inner.base_type_priority()

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        return self.inner.check_and_wrap(arg, BoundTypeVarCtx(self, ctx))


class BoundTypeVarCtx(ExecutionContext):

    def __init__(self, bv: BoundTypeVar, ctx: ExecutionContext):
        self.bv = bv
        self.upper = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (nt, ni) = err.next_type_and_indicator()

        if nt == err.expected and nt == self.bv.inner.describe():
            err.expected = self.bv.describe()

        return self.upper.wrap(err)


class UnboundTypeVar(TypeChecker):

    def __init__(self, typevar: TypeVar):
        self.typevar = typevar

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        return arg

    def describe(self) -> str:
        return str(self.typevar)

    def base_type(self) -> list[Any]:
        return [self.typevar]



#!/usr/bin/env python

from distutils.core import setup
import setuptools
import os
import json

TOP_DIR = os.path.join(os.path.dirname(__file__), '..')
VERSION_FILE = 'VERSION'

def writeVersionFile():
    pkgJson = os.path.join(TOP_DIR, 'package.json')
    if os.path.isfile(pkgJson):
        content = open(pkgJson).read()
        d = json.loads(content)
        version = d['version']
        open(VERSION_FILE, 'w').write(version)

writeVersionFile()

def readVersion():
    return open(VERSION_FILE).read().strip()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='wypp',
      version=readVersion(),
      description='A user-friendly python programming environment for beginners',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Stefan Wehr',
      author_email='stefan.wehr@hs-offenburg.de',
      url='https://github.com/skogsbaer/write-your-python-program',
      package_dir={'wypp': 'src', 'untypy': 'deps/untypy/untypy'},
      packages=['wypp'] + setuptools.find_packages("deps/untypy", exclude=['test', 'test.*']),
      python_requires='>=3.9',
      scripts=['wypp']
      )



import ast
from typing import Callable, List, Optional


class UntypyAstTransformer(ast.NodeTransformer):
    def visit_Module(self, node: ast.Module):
        for i, child in enumerate(node.body):
            if isinstance(child, ast.ImportFrom) and child.module == '__future__':
                continue  # from __future__ import ...
            elif isinstance(child, ast.Expr) and isinstance(child.value, ast.Constant):
                continue  # module docstring
            elif _is_untypy_import(node):
                break
            else:
                node.body.insert(i, ast.Import(names=[ast.alias('untypy', None)]))
                break

        self.generic_visit(node)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef):
        node.decorator_list.insert(0, ast.Attribute(ast.Name("untypy", ast.Load()), "patch", ast.Load()))
        self.generic_visit(node)
        return node

    def visit_ClassDef(self, node: ast.FunctionDef):
        node.decorator_list.insert(0, ast.Attribute(ast.Name("untypy", ast.Load()), "patch", ast.Load()))
        self.generic_visit(node)
        return node

    def visit_Expr(self, node: ast.Expr):
        val = node.value
        if _is_untypy_patch_call(val):
            return ast.Expr(ast.Constant("# untypy.enable()"))
        else:
            self.generic_visit(node)
            return node


class UntypyAstImportTransformer(ast.NodeTransformer):
    def __init__(self, predicate: Callable[[str], bool], module_path: List[str]):
        self.predicate = predicate
        self.module_path = module_path

    def resolve_relative_name(self, name: str, levels: Optional[int]) -> str:
        if levels is None:
            return name
        elif levels == 1:
            return '.'.join(self.module_path + list(name.split('.')))
        else:
            prefix = self.module_path[:-(levels - 1)]
            return '.'.join(prefix + list(name.split('.')))

    def visit_Module(self, node: ast.Module):
        inserts = []
        need_insert_utypy_imp = True
        need_insert_utypy_idx = 0
        for i, child in enumerate(node.body):
            if _is_untypy_import(node):
                need_insert_utypy_imp = False
            elif isinstance(child, ast.ImportFrom) and child.module == '__future__':
                need_insert_utypy_idx += 1
            elif isinstance(child, ast.ImportFrom) and self.predicate(
                    self.resolve_relative_name(child.module, child.level)):
                for alias in child.names:
                    if alias.asname is None:
                        destname = alias.name
                    else:
                        destname = alias.asname
                    # $destname = untypy.wrap_import($destname)
                    call = ast.Call(ast.Attribute(ast.Name("untypy", ast.Load()), "wrap_import", ast.Load()),
                                    [ast.Name(destname, ast.Load())], [])
                    expr = ast.Assign([ast.Name(destname, ast.Store())], call)
                    node.body.insert(i + 1, expr)

            elif isinstance(child, ast.Import):
                for alias in child.names:
                    if self.predicate(alias.name):
                        if alias.asname is None:
                            destname = alias.name
                        else:
                            destname = alias.asname

                        # $destname = untypy.wrap_import($destname)
                        # python ast weirdness
                        def create_attribute(levels: List[str], op=ast.Store()):
                            if len(levels) == 1:
                                return ast.Name(levels[0], ctx=op)
                            else:
                                return ast.Attribute(create_attribute(levels[:-1], ast.Load()), levels[-1], ctx=op)

                        destpath = destname.split(".")
                        call = ast.Call(ast.Attribute(ast.Name("untypy", ast.Load()), "wrap_import", ast.Load()),
                                        [create_attribute(destpath, ast.Load())], [])
                        expr = ast.Assign([create_attribute(destpath)], call)
                        # inserts.append((i + 1, expr)) # insert after
                        node.body.insert(i + 1, expr)

        # TODO BUG: Multiple imports index mix up
        # TODO BUG:
        for (idx, expr) in inserts:
            node.body.insert(idx, expr)

        if need_insert_utypy_imp:
            node.body.insert(need_insert_utypy_idx, ast.Import(names=[ast.alias('untypy', None)]))

        self.generic_visit(node)
        return node

    def visit_Expr(self, node: ast.Expr):
        val = node.value
        if _is_untypy_patch_call(val):
            return ast.Expr(ast.Constant("# untypy.enable()"))
        else:
            self.generic_visit(node)
            return node

def _is_untypy_patch_call(node):
    if isinstance(node, ast.Expr):
        node = node.value

    return (isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == 'untypy'
            and (node.func.attr == 'enable' or node.func.attr == 'enable_on_imports'))


def _is_untypy_import(node):
    return (isinstance(node, ast.Import)
            and len(node.names) == 1
            and isinstance(node.names[0], ast.alias)
            and node.names[0].name == 'untypy')


def did_no_code_run_before_untypy_enable(node: ast.Module) -> bool:
    for child in node.body:
        if isinstance(child, ast.ImportFrom) and child.module == '__future__':
            continue  # from __future__ import ...
        elif _is_untypy_import(child):
            continue
        elif isinstance(child, ast.Expr) and isinstance(child.value, ast.Constant):
            continue  # module docstring
        elif _is_untypy_patch_call(child):
            return True
        else:
            break

    # untypy.enable() was not first. It is still okay if this call does not exist.
    class DidNoCodeRunVisitor(ast.NodeVisitor):
        def __init__(self):
            self.has_untypycall = False

        def visit_Call(self, node):
            if _is_untypy_patch_call(node):
                self.has_untypycall = True
            self.generic_visit(node)
            return node

    visitor = DidNoCodeRunVisitor()
    visitor.visit(node)
    if visitor.has_untypycall:
        return False
    else:
        return True



import abc
from typing import Any, Optional, Callable

from untypy.error import UntypyTypeError
from untypy.impl.protocol import ProtocolChecker
from untypy.impl.wrappedclass import WrappedType
from untypy.interfaces import TypeChecker, TypeCheckerFactory, CreationContext, ExecutionContext


class SimpleFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        ta = type(annotation)
        if ta is type or ta is abc.ABCMeta or annotation is Callable:
            return SimpleChecker(annotation, ctx)
        else:
            return None


class ParentProtocolChecker(ProtocolChecker):
    def protocol_type(self) -> str:
        return "Parent"

def simpleTypeCompat(x: Any, ty: type):
    xTy = type(x)
    return xTy is ty or (ty is float and xTy is int)

class SimpleChecker(TypeChecker):
    annotation: type
    always_wrap: bool = False
    parent_checker: Optional[Callable[[Any, ExecutionContext], Any]]

    def __init__(self, annotation: type, ctx: CreationContext):
        self.annotation = annotation
        self.always_wrap = False

        # use protocol like wrapping only if there are some signatures
        if ctx.should_be_inheritance_checked(annotation):
            if hasattr(annotation, '__patched'):
                p = ParentProtocolChecker(annotation, ctx)
                self.parent_checker = p.check_and_wrap
            else:
                # annotation is from an wrapped import
                t = WrappedType(annotation, ctx)

                def wrap(i, ctx):
                    instance = t.__new__(t)
                    instance._WrappedClassFunction__inner = i
                    instance._WrappedClassFunction__return_ctx = None
                    return instance

                # TODO: Use only on import_wrapped module
                # self.always_wrap = True
                self.parent_checker = wrap
        else:
            self.parent_checker = None


    def may_be_wrapped(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if simpleTypeCompat(arg, self.annotation) and not self.always_wrap:
            return arg
        if isinstance(arg, self.annotation):
            if self.parent_checker is None:
                return arg
            else:
                return self.parent_checker(arg, ctx)
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        return self.annotation.__name__

    def base_type(self) -> Any:
        return [self.annotation]



import inspect
import sys
import typing
from typing import Protocol, Any, Optional, Callable, Union, TypeVar, Dict, Tuple

from untypy.error import UntypyTypeError, UntypyAttributeError, Frame, Location, ResponsibilityType
from untypy.impl.any import SelfChecker
from untypy.interfaces import TypeCheckerFactory, CreationContext, TypeChecker, ExecutionContext, \
    WrappedFunctionContextProvider
from untypy.util import WrappedFunction, ArgumentExecutionContext, ReturnExecutionContext
from untypy.util.condition import FunctionCondition


class ProtocolFactory(TypeCheckerFactory):

    def create_from(self, annotation: Any, ctx: CreationContext) -> Optional[TypeChecker]:
        if isinstance(annotation, type) and Protocol in annotation.mro():
            return ProtocolChecker(annotation, ctx)
        elif hasattr(annotation, '__args__') and hasattr(annotation.__origin__,
                                                         '__mro__') and typing.Protocol in annotation.__origin__.__mro__:
            return ProtocolChecker(annotation, ctx)
        else:
            return None


def _find_bound_typevars(clas: type) -> (type, Dict[TypeVar, Any]):
    if not hasattr(clas, '__args__') or not hasattr(clas, '__origin__'):
        return (clas, dict())
    if not hasattr(clas.__origin__, '__parameters__'):
        return (clas, dict())

    keys = clas.__origin__.__parameters__
    values = clas.__args__

    if len(keys) != len(values):
        raise UntypyAttributeError(f"Some unbound Parameters in {clas.__name__}. "
                                   f"keys={keys} do not match values={values}.",
                                   [Location(
                                       file=inspect.getfile(clas),
                                       line_no=inspect.getsourcelines(clas)[1],
                                       source_line="".join(inspect.getsourcelines(clas)[0]))])
    return (clas.__origin__, dict(zip(keys, values)))


def get_proto_members(proto: type, ctx: CreationContext) -> Dict[
    str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]]:
    blacklist = ['__init__', '__class__', '__delattr__', '__dict__', '__dir__',
                 '__doc__', '__getattribute__', '__getattr__', '__init_subclass__',
                 '__new__', '__setattr__', '__subclasshook__', '__weakref__',
                 '__abstractmethods__', '__class_getitem__']

    member_dict = {}
    for [name, member] in inspect.getmembers(proto):
        if name in blacklist:
            continue

        if inspect.isfunction(member):
            member = WrappedFunction.find_original(member)
            signature = inspect.signature(member)
            annotations = typing.get_type_hints(member, include_extras=True)
            checkers = {}
            for key in signature.parameters:
                if key == 'self':
                    checkers[key] = SelfChecker()
                else:
                    param = signature.parameters[key]
                    if param.annotation is inspect.Parameter.empty:
                        raise ctx.wrap(UntypyAttributeError(
                            f"Missing annotation for argument '{key}' of function {member.__name__} "
                            f"in protocol {proto.__name__}\n"))

                    checker = ctx.find_checker(annotations[key])
                    if checker is None:
                        raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {param.annotation}\n"
                                                            f"for argument '{key}' of function {member.__name__} "
                                                            f"in protocol {proto.__name__}.\n"))
                    checkers[key] = checker

            if signature.return_annotation is inspect.Parameter.empty:
                raise ctx.wrap(UntypyAttributeError(
                    f"Missing annotation for return value of function {member.__name__} "
                    f"in protocol {proto.__name__}. Use 'None' if there is no return value.\n"))
            return_annotation = annotations['return']
            if return_annotation is proto:  # Self as Return Type would led to endless recursion
                return_checker = SimpleInstanceOfChecker(proto, None)
            else:
                return_checker = ctx.find_checker(return_annotation)

            if return_checker is None:
                raise ctx.wrap(UntypyAttributeError(f"\n\tUnsupported type annotation: {signature.return_annotation}\n"
                                                    f"for return value of function {member.__name__} "
                                                    f"in protocol-like {proto.__name__}.\n"))
            fc = None
            if hasattr(member, '__fc'):
                fc = getattr(member, '__fc')
            checkers['return'] = return_checker
            member_dict[name] = (signature, checkers, fc)
    return member_dict


class ProtocolChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext, *, altname : Optional[str] = None):
        (proto, typevars) = _find_bound_typevars(annotation)
        ctx = ctx.with_typevars(typevars)
        members = get_proto_members(proto, ctx)
        self.proto = proto
        self.members = members
        self.typevars = typevars
        self.wrapper_types = dict()
        self.altname = altname

    def may_change_identity(self) -> bool:
        return True

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if hasattr(arg, '_ProtocolWrappedFunction__inner'):
            # no double wrapping
            arg = getattr(arg, '_ProtocolWrappedFunction__inner')

        if type(arg) in self.wrapper_types:
            return self.wrapper_types[type(arg)](arg, ctx)
        else:
            wrapped_type = ProtocolWrapper(self, type(arg), self.members, ctx)
            self.wrapper_types[type(arg)] = wrapped_type
            return wrapped_type(arg, ctx)

    def base_type(self) -> list[Any]:
        # Prevent Classes implementing multiple Protocols in one Union by accident.
        return [Protocol]

    def describe(self) -> str:
        if self.altname is not None:
            return self.altname

        desc = set([])
        for name in self.members:
            (sig, binds, cond) = self.members[name]
            for argname in sig.parameters:
                if isinstance(sig.parameters[argname].annotation, TypeVar):
                    desc.add(binds[argname].describe())
            if isinstance(sig.return_annotation, TypeVar):
                desc.add(binds['return'].describe())
        if len(desc) > 0:
            return f"{self.proto.__name__}[" + (', '.join(desc)) + "]"
        else:
            return f"{self.proto.__name__}"

    def protocol_type(self) -> str:
        return f"protocol"

    def protoname(self):
        return self.describe()


def ProtocolWrapper(protocolchecker: ProtocolChecker, original: type,
                    members: Dict[str, Tuple[inspect.Signature, dict[str, TypeChecker], FunctionCondition]],
                    ctx: ExecutionContext):
    list_of_attr = dict()
    for fnname in members:
        if not hasattr(original, fnname):
            raise ctx.wrap(UntypyTypeError(
                expected=protocolchecker.describe(),
                given=original.__name__
            )).with_note(
                f"Type {original.__name__} does not meet the requirements of protocol {protocolchecker.proto.__name__}. It is missing the function '{fnname}'")

        original_fn = getattr(original, fnname)
        try:
            # fails on built ins - YEAH
            original_fn_signature = inspect.signature(original_fn)
        except:
            original_fn_signature = None

        if hasattr(original_fn, '__wf'):
            original_fn = getattr(original_fn, '__wf')
        (sig, argdict, fc) = members[fnname]

        for param in sig.parameters:
            if original_fn_signature is not None and param not in original_fn_signature.parameters:
                raise ctx.wrap(UntypyTypeError(
                    expected=protocolchecker.describe(),
                    given=original.__name__
                )).with_note(
                    f"Type {original.__name__} does not meet the requirements of protocol {protocolchecker.proto.__name__}. The signature of '{fnname}' does not match. Missing required parameter {param}")

        list_of_attr[fnname] = ProtocolWrappedFunction(original_fn, sig, argdict, protocolchecker, fc).build()

    def constructor(me, inner, ctx):
        me._ProtocolWrappedFunction__inner = inner
        me._ProtocolWrappedFunction__ctx = ctx

    def __getattr__(me, name):
        return getattr(me._ProtocolWrappedFunction__inner, name)

    def __setattr__(me, name, value):
        if name == '_ProtocolWrappedFunction__inner':
            super(type(me), me).__setattr__('_ProtocolWrappedFunction__inner', value)
            return
        if name == '_ProtocolWrappedFunction__ctx':
            super(type(me), me).__setattr__('_ProtocolWrappedFunction__ctx', value)
            return

        return setattr(me._ProtocolWrappedFunction__inner, name, value)

    list_of_attr['__init__'] = constructor
    list_of_attr['__getattr__'] = __getattr__  # allow access of attributes
    list_of_attr['__setattr__'] = __setattr__  # allow access of attributes
    name = f"{protocolchecker.proto.__name__}For{original.__name__}"
    return type(name, (), list_of_attr)


class ProtocolWrappedFunction(WrappedFunction):

    def __init__(self, inner: Union[Callable, WrappedFunction], signature: inspect.Signature,
                 checker: Dict[str, TypeChecker],
                 protocol: ProtocolChecker,
                 fc: FunctionCondition):
        self.inner = inner
        self.signature = signature
        self.checker = checker
        self.protocol = protocol
        self.fc = fc

    def build(self):
        fn = WrappedFunction.find_original(self.inner)

        fn_of_protocol = getattr(self.protocol.proto, fn.__name__)
        if hasattr(fn_of_protocol, '__wf'):
            fn_of_protocol = getattr(fn_of_protocol, '__wf')

        def wrapper(me, *args, **kwargs):
            inner_object = me.__inner
            inner_ctx = me.__ctx

            caller = sys._getframe(1)
            (args, kwargs, bind1) = self.wrap_arguments(lambda n: ArgumentExecutionContext(fn_of_protocol, caller, n),
                                                        (inner_object, *args), kwargs)
            if isinstance(self.inner, WrappedFunction):
                (args, kwargs, bind2) = self.inner.wrap_arguments(lambda n:
                                                                  ProtocolArgumentExecutionContext(self, n,
                                                                                                   inner_object,
                                                                                                   inner_ctx),
                                                                  args, kwargs)
            ret = fn(*args, **kwargs)
            if isinstance(self.inner, WrappedFunction):
                ret = self.inner.wrap_return(ret, bind2, ProtocolReturnExecutionContext(self,
                                                                                        ResponsibilityType.IN,
                                                                                        inner_object,
                                                                                        inner_ctx))
            return self.wrap_return(ret, bind1, ProtocolReturnExecutionContext(self,
                                                                               ResponsibilityType.OUT, inner_object,
                                                                               inner_ctx))

        async def async_wrapper(*args, **kwargs):
            raise AssertionError("Not correctly implemented see wrapper")

        if inspect.iscoroutine(self.inner):
            w = async_wrapper
        else:
            w = wrapper

        setattr(w, '__wrapped__', fn)
        setattr(w, '__name__', fn.__name__)
        setattr(w, '__signature__', self.signature)
        setattr(w, '__wf', self)
        return w

    def get_original(self):
        return self.inner

    def wrap_arguments(self, ctxprv: WrappedFunctionContextProvider, args, kwargs):
        bindings = self.signature.bind(*args, **kwargs)
        bindings.apply_defaults()
        if self.fc is not None:
            self.fc.prehook(bindings, ctxprv)
        for name in bindings.arguments:
            check = self.checker[name]
            ctx = ctxprv(name)
            bindings.arguments[name] = check.check_and_wrap(bindings.arguments[name], ctx)
        return bindings.args, bindings.kwargs, bindings

    def wrap_return(self, ret, bindings, ctx: ExecutionContext):
        check = self.checker['return']
        if self.fc is not None:
            self.fc.posthook(ret, bindings, ctx)
        return check.check_and_wrap(ret, ctx)

    def describe(self) -> str:
        fn = WrappedFunction.find_original(self.inner)
        return f"{fn.__name__}" + str(self.signature)

    def checker_for(self, name: str) -> TypeChecker:
        return self.checker[name]

    def declared(self) -> Location:
        fn = WrappedFunction.find_original(self.inner)
        return WrappedFunction.find_location(getattr(self.protocol.proto, fn.__name__))


class ProtocolReturnExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction, invert: ResponsibilityType, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.invert = invert
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        err = ReturnExecutionContext(self.wf).wrap(err)

        if err.responsibility_type is self.invert:
            return err

        responsable = WrappedFunction.find_location(self.wf)
        (decl, ind) = err.next_type_and_indicator()
        err = err.with_inverted_responsibility_type()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        inner = self.wf.inner
        if isinstance(inner, WrappedFunction):
            err = err.with_note(
                f"The return value of method '{WrappedFunction.find_original(self.wf).__name__}' does violate the {self.wf.protocol.protocol_type()} '{self.wf.protocol.proto.__name__}'.")
            err = err.with_note(
                f"The annotation '{inner.checker_for('return').describe()}' is incompatible with the {self.wf.protocol.protocol_type()}'s annotation '{self.wf.checker_for('return').describe()}'\nwhen checking against the following value:")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_note(
            f"Type '{type(self.me).__name__}' does not implement {self.wf.protocol.protocol_type()} '{self.wf.protocol.protoname()}' correctly.")

        previous_chain = self.ctx.wrap(previous_chain)
        return err.with_previous_chain(previous_chain)


class ProtocolArgumentExecutionContext(ExecutionContext):
    def __init__(self, wf: ProtocolWrappedFunction, arg_name: str, me: Any, ctx: ExecutionContext):
        self.wf = wf
        self.arg_name = arg_name
        self.me = me
        self.ctx = ctx

    def wrap(self, err: UntypyTypeError) -> UntypyTypeError:
        (original_expected, _ind) = err.next_type_and_indicator()
        err = ArgumentExecutionContext(self.wf, None, self.arg_name).wrap(err)

        responsable = WrappedFunction.find_location(self.wf)

        (decl, ind) = err.next_type_and_indicator()
        err = err.with_frame(Frame(
            decl,
            ind,
            declared=self.wf.declared(),
            responsable=responsable
        ))

        err = err.with_note(
            f"The argument '{self.arg_name}' of method '{WrappedFunction.find_original(self.wf).__name__}' violates the {self.wf.protocol.protocol_type()} '{self.wf.protocol.proto.__name__}'.")
        err = err.with_note(
            f"The annotation '{original_expected}' is incompatible with the {self.wf.protocol.protocol_type()}'s annotation '{self.wf.checker_for(self.arg_name).describe()}'\nwhen checking against the following value:")

        previous_chain = UntypyTypeError(
            self.me,
            f"{self.wf.protocol.protoname()}"
        ).with_note(
            f"Type '{type(self.me).__name__}' does not implement {self.wf.protocol.protocol_type()} '{self.wf.protocol.protoname()}' correctly.")

        previous_chain = self.ctx.wrap(previous_chain)
        # err = err.with_inverted_responsibility_type()

        return err.with_previous_chain(previous_chain)


class SimpleInstanceOfChecker(TypeChecker):
    def __init__(self, annotation: type, ctx: CreationContext):
        self.annotation = annotation

    def check_and_wrap(self, arg: Any, ctx: ExecutionContext) -> Any:
        if isinstance(arg, self.annotation):
            return arg
        else:
            raise ctx.wrap(UntypyTypeError(arg, self.describe()))

    def describe(self) -> str:
        return self.annotation.__name__

    def base_type(self) -> Any:
        return [self.annotation]



class IndicatorStr:
    ty: str
    indicator: str

    def __init__(self, ty: str, indicator: str = ""):
        while len(indicator) < len(ty):
            indicator += " "

        self.ty = ty
        self.indicator = indicator

    def __add__(self, other):
        return IndicatorStr(self.ty + other.ty, self.indicator + other.indicator)

    def join(self, lst):
        return IndicatorStr(
            self.ty.join(map(lambda s: s.ty, lst)),
            self.indicator.join(map(lambda s: s.indicator, lst)),
        )
